
/*	Copyright (C) G. Jespersen
	LAV Filters - Copyright (C) 2010-2013 Hendrik Leppkes http://www.1f0.de
	See LAV Licence.txt.*/


#include "stdafx.h"
#include "Khan.h"

#include "Player.h"
#include "math.h"
#include "errors.h"
#include "LAVSplitterSettings.h" // LAV Settings interfaces
#include "LAVVideoSettings.h"
#include "LAVAudioSettings.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Note - from dsound.h @ Program Files\Microsoft Visual Studio\VC98\Include
// For vs 2005
#define DSBVOLUME_MIN               -10000	
#define DSBVOLUME_MAX               0


CPlayer::CPlayer()
{
	// Initial Video playback.
	m_pGraph		= NULL;
	m_pControl		= NULL;
	m_pEvent		= NULL;
	m_pSeeking		= NULL;
	m_pVidWin		= NULL;
	m_pBaseVid		= NULL;
	m_pBaseAudio	= NULL;
	m_redrawVid		= FALSE;
	m_muted			= FALSE;
	m_curVolPercent = 50;
	m_fsCurrent		= State_Stopped; // State_Stopped = 0, State_Paused, State_Running
	m_rtTotalTime	= 0;
	m_hWnd			= NULL;			 // playback window, callback, MsgBox
	m_nPlays		= 0;			 // debug

	// LAV Filters.
	m_pLavSrc		= NULL;
	m_pLavVid		= NULL;
	m_pLavAud		= NULL;

	// Get current directory.
	TCHAR szPath[1024];
	::GetCurrentDirectory(sizeof (szPath) / sizeof(TCHAR), szPath);
	m_curDirectory = szPath;

	// Set LAV DLL Paths. Note appending "_1" to files to distinguish
	// from any others on users system.
	//
	// 05/15 upgrade to LAV 0.6.5
	CString lavDir = _T("LAV");
	m_pathLavSrc.Format(_T("%s\\%s\\LAVSplitter_1.ax"), m_curDirectory, lavDir); 
	m_pathLavVid.Format(_T("%s\\%s\\LAVVideo_1.ax"), m_curDirectory, lavDir); 
	m_pathLavAud.Format(_T("%s\\%s\\LAVAudio_1.ax"), m_curDirectory, lavDir); 
}

CPlayer::~CPlayer()
{
	TRACE(_T("In CPlayer destructor.\n"));

	// Clean up. No - see CDoc::OnCloseDocument()
	//CleanPlayback();

	m_hWnd = NULL;
}


// 06/07/16 - Use LAV 0.68.1
// Upgrade to support HTTPS:
//	https://cdn.kastatic.org/KA-youtube-converted/l-QFT7XNeb4.mp4/l-QFT7XNeb4.mp4
//	http://cdn.kastatic.org/KA-youtube-converted/l-QFT7XNeb4.mp4/l-QFT7XNeb4.mp4
//
// Build graph manually with LAV Filters, except for rendering.
// Create LAV Filters using IBaseFilter ptrs.
// Create Video and Audio stream handlers only as needed by media.
// Demo code for LAV Settings interfaces.
// 
// Note memory cleans to about 20mb after stopping Test/PlayMp4Lav,
// per Process Explorer.
//
// Called from:
//	CDataView::StartVideoPlayback() for disk and
//	CDataView::PlayVideoStream() for URL. 12/13/14
BOOL CPlayer::PlayVideoLav(const CString & path, HWND hwnd, CRect rect) 
{
	// problem vids - ui freeze while caching - 5/15 - LAV 0.6.5 no fix:
	// Bernini, Ecstasy of Saint Teresa
	//	http://fastly.kastatic.org/KA-youtube-converted/RKcJvjP9zgY.mp4/RKcJvjP9zgY.mp4
	//	1280 x 720
	//				
	// Apollonius, Seated Boxer
	//	http://fastly.kastatic.org/KA-youtube-converted/FvsSPJoJB3k.mp4/FvsSPJoJB3k.mp4				 
	//	1280 x 720
	//
	// ok
	// http://fastly.kastatic.org/KA-youtube-converted/-bE-Pydad7U.mp4/-bE-Pydad7U.mp4
	// http://fastly.kastatic.org/KA-youtube-converted/BCH1Gre3Mg0.mp4/BCH1Gre3Mg0.mp4
	//	1280 x 720

	TRACE(_T("In CPlayer::PlayVideoLav.\n"));

	IPin * p1 = 0;
	IPin * p2 = 0;
	int ref = 0;
	HRESULT hr;

	// Set hwnd
	m_hWnd = hwnd;

	// Clean any currently playing - needs fixing if used.
	if(m_fsCurrent != State_Stopped)
	{
		if(!CleanPlayback())
			return FALSE;
	}

	////////////////////////////////////
	// Create the filter graph manager.
	hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, 
						  IID_IGraphBuilder, (void **)&m_pGraph);
	if(FAILED(hr))
		return CleanPlayback("ERROR - Could not create the Filter Graph Manager.");

	TRACE(_T("In PlayVideoLav CoCreateInstance done.\n"));

	///////////////////////////////////////////////////
	// Set a dbg Log - file created only if a failure.
	#ifdef _DEBUG
	CString logPath = m_curDirectory;
	logPath += _T("\\DxLog.txt");
	HANDLE hFile = CreateFile(logPath, GENERIC_WRITE, 0, NULL, 
							  CREATE_ALWAYS,
							  FILE_ATTRIBUTE_NORMAL, NULL);
	hr = m_pGraph->SetLogFile((unsigned long)hFile);
	if(FAILED(hr))
		return CleanPlayback("ERROR - SetLogFile failed.");
	#endif

	//////////////////////////////////////////
	// Create the Lav Source Splitter filter.
	CLSID clsidLavSrc = { 0xB98D13E7, 0x55DB, 0x4385, 0xA3, 0x3D, 0x09, 0xFD, 0x1B, 0xA2, 0x63, 0x38 };
	hr = CreateObjectFromPath(m_pathLavSrc, clsidLavSrc, &m_pLavSrc);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - CreateObjectFromPath2 failed - clsidLavSrc"));

	TRACE(_T("In PlayVideoLav Lav Source Splitter clsidLavSrc done.\n"));

	// Set default Lav Source Splitter settings.
	// Note: sample code for LAV settings interfaces, not necessary.
	ILAVFSettings * pSourceSettings;
	hr = m_pLavSrc->QueryInterface(IID_ILAVFSettings, (void **)&pSourceSettings);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - ILAVFSettings not supported"));

	hr = pSourceSettings->SetRuntimeConfig(TRUE); // Set to default. Changes are not saved.
	pSourceSettings->Release();

	TRACE(_T("In PlayVideoLav Lav settings done.\n"));

	// Get IFileSourceFilter interface.
	IFileSourceFilter * pFileSource;
	hr = m_pLavSrc->QueryInterface(IID_IFileSourceFilter,(void **)&pFileSource);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - QueryInterface failed - IID_IFileSourceFilter"));

	// dbg
//	DWORD start = GetTickCount();

	// Load the media via Lav Source Splitter - disk or url.
	//
	// LAV calls FFMPEG avformat_open_input() via CLAVFDemuxer::OpenInputStream - see:
	//	 http://libav-users.943685.n4.nabble.com/Is-av-open-input-file-blocking-normal-td3312316.html
	//	 http://libav-users.943685.n4.nabble.com/timeout-on-av-read-frame-td3306682.html
	//	 http://plagatux.es/2012/08/skipping-blocking-libav-functions-with-interruption-callbacks/
	//
	// Note can have big delays here with some fastly URLs and stream-on-miss:
	//	 188ms disk, 1172 cached, 20297 not cahed.
	//	 See: https://social.msdn.microsoft.com/Forums/windowsdesktop/en-US/7e592373-4e35-42ef-8f9a-4d595e77e243/ifilesourcefilter-asynchronous-operation?forum=windowsdirectshowdevelopment
	//	 see Q224318
	//	 see: https://msdn.microsoft.com/en-us/library/windows/desktop/ms644928%28v=vs.85%29.aspx
	hr = pFileSource->Load(path, NULL);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - Load failed - m_pLavSrc"));
	
	TRACE(_T("PlayVideoLav media loaded done.\n"));
	// dbg
//	TRACE(_T("In PlayVideoLav Load() elapsed time = %d ms ret = %d\n"), GetTickCount() - start, hr);
	
	// Add the Lav Source Splitter filter.
	hr = m_pGraph->AddFilter(m_pLavSrc, L"LAV Splitter Source");
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - AddFilter failed - m_pLavSrc"));
	pFileSource->Release();

	TRACE(_T("PlayVideoLav add Lav Source Splitter done.\n"));

	//////////////////////////////////////
	// Check for Video and Audio streams.
	BOOL haveAudio = FALSE, haveVideo = FALSE;
	hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Video, &p1);
	if(SUCCEEDED(hr))
	{
		haveVideo = TRUE;
		ref = p1->Release(); // 3
	}
	hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Audio, &p2);
	if(SUCCEEDED(hr))
	{
		haveAudio = TRUE;
		ref = p2->Release(); // 3
	}
	if(!haveAudio && !haveVideo) 
		return CleanPlayback("ERROR - Media type not supported.");

	//////////////////////
	// If a Video stream.
	if(haveVideo)
	{
		// Create the Lav Video Decoder filter.
		CLSID clsidLavVid = { 0xEE30215D, 0x164F, 0x4A92, 0xA4, 0xEB, 0x9D, 0x4C, 0x13, 0x39, 0x0F, 0x9F };
		hr = CreateObjectFromPath(m_pathLavVid, clsidLavVid, &m_pLavVid);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - CreateObjectFromPath failed - clsidLavVid"));

		// Set default Lav Video Decoder settings.
		ILAVVideoSettings * pVideoSettings;
		hr = m_pLavVid->QueryInterface(IID_ILAVVideoSettings, (void **)&pVideoSettings);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - ILAVVideoSettings not supported"));
		hr = pVideoSettings->SetRuntimeConfig(TRUE); // Set to default. Changes are not saved.
		pVideoSettings->Release();

		// Add the LAV decoder.
		hr = m_pGraph->AddFilter(m_pLavVid, L"LAV Video Decoder");
		if(FAILED(hr))
			return CleanPlayback("ERROR - AddFilter failed - LAV Video Decoder.");

		// Connect Lav Splitter Source Video to Lav Video Decoder
		hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Video, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Lav Source Splitter Video.");
		hr = GetPin(m_pLavVid, PINDIR_INPUT, MEDIATYPE_Video, &p2);
		if(FAILED(hr))
		{
			p1->Release();
			return CleanPlayback("ERROR - GetPin - In Lav Video Decoder.");
		}
		hr = p1->Connect(p2, NULL);
		ref = p1->Release();
		ref = p2->Release();
		if(FAILED(hr))
			return CleanPlayback("ERROR - Connect failed - Lav Source to Video.");

		// Find and render the LAV Video decoder output pin. 
		hr = GetPin(m_pLavVid, PINDIR_OUTPUT, MEDIATYPE_Video, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Video Decoder.");
		hr = m_pGraph->Render(p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - Render video failed.");
		ref = p1->Release();

		// Set Video interfaces.
		hr = m_pGraph->QueryInterface(IID_IVideoWindow, (void **)&m_pVidWin);
		if(FAILED(hr))
			return CleanPlayback("ERROR - IVideoWindow not supported.");

		hr = m_pGraph->QueryInterface(IID_IBasicVideo, (void **)&m_pBaseVid);
		if(FAILED(hr))
			return CleanPlayback("ERROR - IBasicVideo not supported.");

		// Setup the video window to use our window handle
		hr = m_pVidWin->put_Owner((OAHWND)m_hWnd);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_Owner failed.");

		hr = m_pVidWin->put_MessageDrain((OAHWND)m_hWnd);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_MessageDrain failed.");

		// Set the style of the video window
		hr = m_pVidWin->put_WindowStyle(WS_CHILD | WS_CLIPSIBLINGS);// | WS_CLIPCHILDREN);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_WindowStyle failed.");

		// Set video rect size to native vid scale.
		// Function cleans up if an error.
		if(!SetVideoRect(rect))
			return FALSE;	
	}

	///////////////////////
	// If an Audio stream.
	if(haveAudio)
	{
		// Create the Lav Audio Decoder filter.
		CLSID clsidLavAud = { 0xE8E73B6B, 0x4CB3, 0x44A4, 0xBE, 0x99, 0x4F, 0x7B, 0xCB, 0x96, 0xE4, 0x91 };
		hr = CreateObjectFromPath(m_pathLavAud, clsidLavAud, &m_pLavAud);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - CreateObjectFromPath failed - clsidLavAud"));

		// Set default Lav Audio Decoder settings.
		ILAVAudioSettings * pAudioSettings;
		hr = m_pLavAud->QueryInterface(IID_ILAVAudioSettings, (void **)&pAudioSettings);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - ILAVAudioSettings not supported"));
		hr = pAudioSettings->SetRuntimeConfig(TRUE); // Set to default. Changes are not saved.
		pAudioSettings->Release();

		// Add the Lav Audio Decoder filter.
		hr = m_pGraph->AddFilter(m_pLavAud, L"LAV Audio Decoder");
		if(FAILED(hr))
			return CleanPlayback("ERROR - AddFilter failed - LAV Audio Decoder.");

		// Connect Lav Splitter Source Audio to Lav Audio Decoder
		hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Audio, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Lav Source Splitter Audio.");
		hr = GetPin(m_pLavAud, PINDIR_INPUT, MEDIATYPE_Audio, &p2);
		if(FAILED(hr))
		{
			p1->Release();
			return CleanPlayback("ERROR - GetPin - In Lav Audio Decoder.");
		}
		hr = p1->Connect(p2, NULL);
		ref = p1->Release();
		ref = p2->Release();
		if(FAILED(hr))
			return CleanPlayback("ERROR - Connect failed - Lav Source to Audio.");

		// Find and render the LAV Audio decoder output pin.
		hr = GetPin(m_pLavAud, PINDIR_OUTPUT, MEDIATYPE_Audio, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Audio Decoder.");
		hr = m_pGraph->Render(p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - Render audio failed.");
		ref = p1->Release();
		
		// Set Audio interfaces.
		hr = m_pGraph->QueryInterface(IID_IBasicAudio, (void **)&m_pBaseAudio);
		if(FAILED(hr))
			return CleanPlayback("ERROR - IBasicAudio not supported.");

		// Set volume to normal - 50%.
		LONG norm = DxVolFromPercent(50);
		hr = m_pBaseAudio->put_Volume(norm);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_Volume failed.");
	}

	//////////////////////////
	// Get common interfaces.
	hr = m_pGraph->QueryInterface(IID_IMediaControl, (void **)&m_pControl);
	if(FAILED(hr))
		return CleanPlayback("ERROR - IMediaControl not supported.");

	hr = m_pGraph->QueryInterface(IID_IMediaEventEx, (void **)&m_pEvent); // IID_IMediaEvent
	if(FAILED(hr))
		return CleanPlayback("ERROR - IMediaEventEx not supported.");

	hr = m_pGraph->QueryInterface(IID_IMediaSeeking, (void **)&m_pSeeking);
	if(FAILED(hr))
		return CleanPlayback("ERROR - IMediaSeeking not supported.");

	// Have the graph signal event via window callbacks to the m_hWnd.
	// See GetVideoEvent().
	hr = m_pEvent->SetNotifyWindow((OAHWND)m_hWnd, WM_USER_VIDEOEVENT, 0);
	if(FAILED(hr))
		return CleanPlayback("ERROR - SetNotifyWindow failed.");

	//////////////////
	// Run the graph.
	// Note CPlayerView::OnVideoEvent() will be called at finish.
	TRACE(_T("Calling m_pControl Run\n"));

	hr = m_pControl->Run();

	TRACE(_T("m_pControl Run returned %d\n"), hr);
	if(FAILED(hr))
	{
		return CleanPlayback("ERROR - Run failed.");
	}

	// Check state - have not seen it fail.
	FILTER_STATE fs;
	TRACE(_T("PlayVideoLav calling GetState.\n"));
	hr = m_pControl->GetState(1000, (OAFilterState*)&fs);
	if(FAILED(hr)) 
	{
		TRACE(_T("ERROR PlayVideoLav GetState failed.\n"));
		return CleanPlayback("ERROR - GetState failed.");
	}
	if(fs != State_Running)
	{
		return CleanPlayback("ERROR - not State_Running.");
	}
	TRACE(_T("PlayVideoLav State_Running OK.\n"));

	// Set state
	m_fsCurrent = State_Running;

	// Prepare seeking.
	PrepareSeeking();

	// OK
	return TRUE;
}

/*
// pre fix fastly
// Build graph manually with LAV Filters, except for rendering.
// Create LAV Filters using IBaseFilter ptrs.
// Create Video and Audio stream handlers only as needed by media.
// Demo code for LAV Settings interfaces.
// 
// Note memory cleans to about 20mb after stopping Test/PlayMp4Lav,
// per Process Explorer.
//
// Called from:
//	CDataView::StartVideoPlayback() for disk and
//	CDataView::PlayVideoStream() for URL. 12/13/14
BOOL CPlayer::PlayVideoLav(const CString & path, HWND hwnd, CRect rect) 
{
	TRACE(_T("In CPlayer::PlayVideoLav.\n"));

	IPin * p1 = 0;
	IPin * p2 = 0;
	int ref = 0;
	HRESULT hr;

	// Set hwnd
	m_hWnd = hwnd;

	// Clean any currently playing - needs fixing if used.
	if(m_fsCurrent != State_Stopped)
	{
		if(!CleanPlayback())
			return FALSE;
	}

	////////////////////////////////////
	// Create the filter graph manager.
	hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, 
						  IID_IGraphBuilder, (void **)&m_pGraph);
	if(FAILED(hr))
		return CleanPlayback("ERROR - Could not create the Filter Graph Manager.");

	///////////////////////////////////////////////////
	// Set a dbg Log - file created only if a failure.
	#ifdef _DEBUG
	CString logPath = m_curDirectory;
	logPath += _T("\\DxLog.txt");
	HANDLE hFile = CreateFile(logPath, GENERIC_WRITE, 0, NULL, 
							  CREATE_ALWAYS,
							  FILE_ATTRIBUTE_NORMAL, NULL);
	hr = m_pGraph->SetLogFile((unsigned long)hFile);
	if(FAILED(hr))
		return CleanPlayback("ERROR - SetLogFile failed.");
	#endif

	//////////////////////////////////////////
	// Create the Lav Source Splitter filter.
	CLSID clsidLavSrc = { 0xB98D13E7, 0x55DB, 0x4385, 0xA3, 0x3D, 0x09, 0xFD, 0x1B, 0xA2, 0x63, 0x38 };
	hr = CreateObjectFromPath(m_pathLavSrc, clsidLavSrc, &m_pLavSrc);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - CreateObjectFromPath2 failed - clsidLavSrc"));

	// Set default Lav Source Splitter settings.
	// Note: sample code for LAV settings interfaces, not necessary.
	ILAVFSettings * pSourceSettings;
	hr = m_pLavSrc->QueryInterface(IID_ILAVFSettings, (void **)&pSourceSettings);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - ILAVFSettings not supported"));
	hr = pSourceSettings->SetRuntimeConfig(TRUE); // Set to default. Changes are not saved.
	DWORD sz = pSourceSettings->GetMaxQueueMemSize();
	pSourceSettings->Release();

	// Load the media.
	IFileSourceFilter * pFileSource;
	hr = m_pLavSrc->QueryInterface(IID_IFileSourceFilter,(void **)&pFileSource);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - QueryInterface failed - IID_IFileSourceFilter"));
	hr = pFileSource->Load(path, NULL);
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - Load failed - m_pLavSrc"));

	// Add the Lav Source Splitter filter.
	hr = m_pGraph->AddFilter(m_pLavSrc, L"LAV Splitter Source");
	if(FAILED(hr))
		return CleanPlayback(_T("ERROR - AddFilter failed - m_pLavSrc"));
	pFileSource->Release();

	//////////////////////////////////////
	// Check for Video and Audio streams.
	BOOL haveAudio = FALSE, haveVideo = FALSE;
	hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Video, &p1);
	if(SUCCEEDED(hr))
	{
		haveVideo = TRUE;
		ref = p1->Release(); // 3
	}
	hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Audio, &p2);
	if(SUCCEEDED(hr))
	{
		haveAudio = TRUE;
		ref = p2->Release(); // 3
	}
	if(!haveAudio && !haveVideo) 
		return CleanPlayback("ERROR - Media type not supported.");

	//////////////////////
	// If a Video stream.
	if(haveVideo)
	{
		// Create the Lav Video Decoder filter.
		CLSID clsidLavVid = { 0xEE30215D, 0x164F, 0x4A92, 0xA4, 0xEB, 0x9D, 0x4C, 0x13, 0x39, 0x0F, 0x9F };
		hr = CreateObjectFromPath(m_pathLavVid, clsidLavVid, &m_pLavVid);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - CreateObjectFromPath failed - clsidLavVid"));

		// Set default Lav Video Decoder settings.
		ILAVVideoSettings * pVideoSettings;
		hr = m_pLavVid->QueryInterface(IID_ILAVVideoSettings, (void **)&pVideoSettings);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - ILAVVideoSettings not supported"));
		hr = pVideoSettings->SetRuntimeConfig(TRUE); // Set to default. Changes are not saved.
		pVideoSettings->Release();

		// Add the LAV decoder.
		hr = m_pGraph->AddFilter(m_pLavVid, L"LAV Video Decoder");
		if(FAILED(hr))
			return CleanPlayback("ERROR - AddFilter failed - LAV Video Decoder.");

		// Connect Lav Splitter Source Video to Lav Video Decoder
		hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Video, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Lav Source Splitter Video.");
		hr = GetPin(m_pLavVid, PINDIR_INPUT, MEDIATYPE_Video, &p2);
		if(FAILED(hr))
		{
			p1->Release();
			return CleanPlayback("ERROR - GetPin - In Lav Video Decoder.");
		}
		hr = p1->Connect(p2, NULL);
		ref = p1->Release();
		ref = p2->Release();
		if(FAILED(hr))
			return CleanPlayback("ERROR - Connect failed - Lav Source to Video.");

		// Find and render the LAV Video decoder output pin. 
		hr = GetPin(m_pLavVid, PINDIR_OUTPUT, MEDIATYPE_Video, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Video Decoder.");
		hr = m_pGraph->Render(p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - Render video failed.");
		ref = p1->Release();

		// Set Video interfaces.
		hr = m_pGraph->QueryInterface(IID_IVideoWindow, (void **)&m_pVidWin);
		if(FAILED(hr))
			return CleanPlayback("ERROR - IVideoWindow not supported.");

		hr = m_pGraph->QueryInterface(IID_IBasicVideo, (void **)&m_pBaseVid);
		if(FAILED(hr))
			return CleanPlayback("ERROR - IBasicVideo not supported.");

		// Setup the video window to use our window handle
		hr = m_pVidWin->put_Owner((OAHWND)m_hWnd);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_Owner failed.");

		hr = m_pVidWin->put_MessageDrain((OAHWND)m_hWnd);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_MessageDrain failed.");

		// Set the style of the video window
		hr = m_pVidWin->put_WindowStyle(WS_CHILD | WS_CLIPSIBLINGS);// | WS_CLIPCHILDREN);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_WindowStyle failed.");

		// Set video rect size to native vid scale.
		// Function cleans up if an error.
		if(!SetVideoRect(rect))
			return FALSE;	
	}

	///////////////////////
	// If an Audio stream.
	if(haveAudio)
	{
		// Create the Lav Audio Decoder filter.
		CLSID clsidLavAud = { 0xE8E73B6B, 0x4CB3, 0x44A4, 0xBE, 0x99, 0x4F, 0x7B, 0xCB, 0x96, 0xE4, 0x91 };
		hr = CreateObjectFromPath(m_pathLavAud, clsidLavAud, &m_pLavAud);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - CreateObjectFromPath failed - clsidLavAud"));

		// Set default Lav Audio Decoder settings.
		ILAVAudioSettings * pAudioSettings;
		hr = m_pLavAud->QueryInterface(IID_ILAVAudioSettings, (void **)&pAudioSettings);
		if(FAILED(hr))
			return CleanPlayback(_T("ERROR - ILAVAudioSettings not supported"));
		hr = pAudioSettings->SetRuntimeConfig(TRUE); // Set to default. Changes are not saved.
		pAudioSettings->Release();

		// Add the Lav Audio Decoder filter.
		hr = m_pGraph->AddFilter(m_pLavAud, L"LAV Audio Decoder");
		if(FAILED(hr))
			return CleanPlayback("ERROR - AddFilter failed - LAV Audio Decoder.");

		// Connect Lav Splitter Source Audio to Lav Audio Decoder
		hr = GetPin(m_pLavSrc, PINDIR_OUTPUT, MEDIATYPE_Audio, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Lav Source Splitter Audio.");
		hr = GetPin(m_pLavAud, PINDIR_INPUT, MEDIATYPE_Audio, &p2);
		if(FAILED(hr))
		{
			p1->Release();
			return CleanPlayback("ERROR - GetPin - In Lav Audio Decoder.");
		}
		hr = p1->Connect(p2, NULL);
		ref = p1->Release();
		ref = p2->Release();
		if(FAILED(hr))
			return CleanPlayback("ERROR - Connect failed - Lav Source to Audio.");

		// Find and render the LAV Audio decoder output pin.
		hr = GetPin(m_pLavAud, PINDIR_OUTPUT, MEDIATYPE_Audio, &p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - GetPin - Out Audio Decoder.");
		hr = m_pGraph->Render(p1);
		if(FAILED(hr))
			return CleanPlayback("ERROR - Render audio failed.");
		ref = p1->Release();
		
		// Set Audio interfaces.
		hr = m_pGraph->QueryInterface(IID_IBasicAudio, (void **)&m_pBaseAudio);
		if(FAILED(hr))
			return CleanPlayback("ERROR - IBasicAudio not supported.");

		// Set volume to normal - 50%.
		LONG norm = DxVolFromPercent(50);
		hr = m_pBaseAudio->put_Volume(norm);
		if(FAILED(hr))
			return CleanPlayback("ERROR - put_Volume failed.");
	}

	//////////////////////////
	// Get common interfaces.
	hr = m_pGraph->QueryInterface(IID_IMediaControl, (void **)&m_pControl);
	if(FAILED(hr))
		return CleanPlayback("ERROR - IMediaControl not supported.");

	hr = m_pGraph->QueryInterface(IID_IMediaEventEx, (void **)&m_pEvent); // IID_IMediaEvent
	if(FAILED(hr))
		return CleanPlayback("ERROR - IMediaEventEx not supported.");

	hr = m_pGraph->QueryInterface(IID_IMediaSeeking, (void **)&m_pSeeking);
	if(FAILED(hr))
		return CleanPlayback("ERROR - IMediaSeeking not supported.");

	// Have the graph signal event via window callbacks to the m_hWnd.
	// See GetVideoEvent().
	hr = m_pEvent->SetNotifyWindow((OAHWND)m_hWnd, WM_USER_VIDEOEVENT, 0);
	if(FAILED(hr))
		return CleanPlayback("ERROR - SetNotifyWindow failed.");

	//////////////////
	// Run the graph.
	// Note CPlayerView::OnVideoEvent() will be called at finish.
	TRACE(_T("Calling m_pControl Run\n"));

	hr = m_pControl->Run();

	TRACE(_T("m_pControl Run returned %d\n"), hr);
	if(FAILED(hr))
		return CleanPlayback("ERROR - Run failed.");

	// Check state - have not seen it fail.
	FILTER_STATE fs;
	TRACE(_T("PlayVideoLav calling GetState.\n"));
	hr = m_pControl->GetState(1000, (OAFilterState*)&fs);
	if(FAILED(hr)) 
	{
		TRACE(_T("ERROR PlayVideoLav GetState failed.\n"));
		return CleanPlayback("ERROR - GetState failed.");// handle error
	}
	if(fs != State_Running)
		return CleanPlayback("ERROR - not State_Running.");// handle error
	TRACE(_T("PlayVideoLav State_Running OK.\n"));

	// Set state
	m_fsCurrent = State_Running;

	// Prepare seeking.
	PrepareSeeking();

	// OK
	return TRUE;
}*/

// Get IBaseFilter ptr from its path and clsid.
// Release() must be called on ptr after use.
// Hacked from: 
//  http://www.gdcl.co.uk/2011/June/UnregisteredFilters.htm  and
//	https://raw.github.com/william0wang/dshownative/master/dump.cpp
HRESULT CPlayer::CreateObjectFromPath(CString path, REFCLSID clsid, 
										  IBaseFilter ** pBaseFilter)
{
	DllGetBaseFilterFunc pFunc;
	HRESULT hr;

	// Load the target DLL directly.
	HMODULE lib = AfxLoadLibrary(path); // or LoadLibrary
	if(!lib)
	{
		return HRESULT_FROM_WIN32(GetLastError());
	}

	// The entry point is an exported function.
	pFunc = (DllGetBaseFilterFunc)GetProcAddress(lib,"DllGetClassObject");
	if(pFunc == NULL)
	{
		return HRESULT_FROM_WIN32(GetLastError());
	}

	// Create a class factory.
	IClassFactory * pCf = NULL;
	hr = pFunc(clsid, IID_IClassFactory, (void**)&pCf);
	if(FAILED(hr))
		return hr;

	// Ask the class factory to create an instance.
	IUnknown * pUnk = NULL;
	hr = pCf->CreateInstance(NULL, IID_IUnknown, (void **)&pUnk);
	if(FAILED(hr))
		return hr;
	pCf->Release();

	// Get the IBaseFilter ptr.
	hr = pUnk->QueryInterface(IID_IBaseFilter, (void **)pBaseFilter);
	if(FAILED(hr))
		return hr;
	pUnk->Release();

	// Ensure.
	if(!pBaseFilter)
		return S_FALSE;

	return S_OK;
}

// From msdn - Working with Pin Categories
HRESULT CPlayer::GetPin(IBaseFilter * pF,		// Pointer to the filter to search.
						PIN_DIRECTION PinDir,	// Direction of the pin.
						const GUID& Category,	// Pin category.
						IPin **ppPin)			// Receives a pointer to the pin.
{
	if(!pF || !ppPin) 
		return E_POINTER;

	*ppPin = 0;
	HRESULT hr;
	IEnumPins *pEnum = 0;
	if (SUCCEEDED(pF->EnumPins(&pEnum)))
	{
		IPin *pPin = 0;
		while (hr = pEnum->Next(1, &pPin, 0), hr == S_OK)
		{
			PIN_DIRECTION ThisPinDir;
			hr = pPin->QueryDirection(&ThisPinDir);
			if (FAILED(hr))
			{
				pPin->Release();
				pEnum->Release();
				return E_UNEXPECTED; // Something strange happened.
			}
			if( (ThisPinDir == PinDir) && PinMatchesCategory(pPin, Category, PinDir) )
			{
				*ppPin = pPin;  // Caller must release the interface.
				pEnum->Release();
				return S_OK;
			}
			pPin->Release();
		}
		pEnum->Release();
    }
    // No pin matches.
    return E_FAIL;
}

// From msdn - Working with Pin Categories - but with AM_MEDIA_TYPE.
BOOL CPlayer::PinMatchesCategory(IPin * pPin, const GUID & Category, PIN_DIRECTION PinDir)
{
    if(pPin == NULL) 
		return E_POINTER;

	// Check for a major media type match.
	BOOL bFound = FALSE;
	int nMedTypes = 0;
	IEnumMediaTypes * ppEnum;
	AM_MEDIA_TYPE	* amType = NULL;
	int ref = 0;
	
	HRESULT hr = pPin->EnumMediaTypes(&ppEnum);
	if(SUCCEEDED(hr))
	{
		ULONG fetched;
		while(ppEnum->Next(1, &amType, &fetched) == S_OK)// fails with lav input
		{
			nMedTypes ++;
			if(amType->majortype == Category)
			{
				bFound = TRUE;
				ref = ppEnum->Release();
				ppEnum = NULL;

				// Delete a heap allocated AM_MEDIA_TYPE structure
				UtilDeleteMediaType(amType);
				amType = NULL;
				break;
			}
			// Delete a heap allocated AM_MEDIA_TYPE structure
			UtilDeleteMediaType(amType);
			amType = NULL;
		}
		// Release IEnumMediaTypes
		if(ppEnum)
			ref = ppEnum->Release();
	}

	// If no media types available and looking for an input pin, 
	// go for it as some inputs will not report a media type.
	if(!nMedTypes && !bFound && PinDir == PINDIR_INPUT)
		bFound = TRUE;

	return bFound;
}

// From DShowUtil
void CPlayer::UtilDeleteMediaType(AM_MEDIA_TYPE *pmt)
{
    // Allow NULL pointers for coding simplicity
    if (pmt == NULL) 
	{
        return;
    }

    // Free media type's format data
    if (pmt->cbFormat != 0) 
    {
        CoTaskMemFree((PVOID)pmt->pbFormat);

        // Strictly unnecessary but tidier
        pmt->cbFormat = 0;
        pmt->pbFormat = NULL;
    }

    // Release interface
    if (pmt->pUnk != NULL) 
    {
        pmt->pUnk->Release();
        pmt->pUnk = NULL;
    }

    // Free media type
    CoTaskMemFree((PVOID)pmt);
}

// Clean up video playback error and show message.
// Always return FALSE.
BOOL CPlayer::CleanPlayback(const CString & errMsg)
{
	CleanPlayback();
	MessageBox(m_hWnd, errMsg, _T("Khan Desktop"), MB_OK);

	// Error 
	return FALSE;
}

// Clean up video playback error and show msg param, dx msg, and hresult.
// Always return FALSE.
BOOL CPlayer::CleanPlayback(const CString & errMsg, int hr)
{
	CleanPlayback();

	CString am, msg;
	TCHAR temp[1024];

	// Retrieve the error message for the hr param return code.
	AMGetErrorText(hr, temp, sizeof(TCHAR) * 1024);
	am = temp;

	// Display.
	msg.Format(_T("%s\n%s\nCode %d"), errMsg, am, hr);
	MessageBox(m_hWnd, msg, _T("Khan Desktop"), MB_OK);

	// Error 
	return FALSE;
}

// Clean up video playback.
// Return TRUE unless cleanup fails.
BOOL CPlayer::CleanPlayback()
{
	TRACE(_T("In CPlayer::CleanPlayback.\n"));

	HRESULT hr;
	BOOL ret = TRUE;
	int ref = 0;

	// Stop the graph (first).
	if(m_pControl)
	{
		hr = m_pControl->Stop();
		if(FAILED(hr))
		{
			MessageBox(m_hWnd, 
					   _T("ERROR CleanPlayback - Stop failed."), 
					   _T("Khan Desktop"), MB_OK);
			ret = FALSE;
		}
	}

	// Use the WaitForCompletion method to wait for the stop operation 
	// to complete. See msdn Using Event Notification.
	long evCode;
	if(m_pEvent)
	{
		hr = m_pEvent->WaitForCompletion(INFINITE, &evCode);
		if(FAILED(hr))
		{
			// Fails if the graph is not in or transitioning into a running state.
		}
	}

	// Clean video window.
	if(m_pVidWin)
	{
		hr = m_pVidWin->put_Visible(OAFALSE);
		if(FAILED(hr))
		{
			MessageBox(m_hWnd, 
					   _T("ERROR CleanPlayback - put_Visable failed."), 
					   _T("Khan Desktop"), MB_OK);
			ret = FALSE;
		}

		hr = m_pVidWin->put_Owner(NULL);
		if(FAILED(hr))
		{
			MessageBox(m_hWnd, 
					   _T("ERROR CleanPlayback - put_Owner failed."), 
					   _T("Khan Desktop"), MB_OK);
			ret = FALSE;
		}
	}

	// Debug - show the filters actually used.
	#ifdef _DEBUG
	EnumFilters();
	#endif*

	// Release interfaces.
	if(m_pSeeking)
		ref = m_pSeeking->Release();	
	if(m_pEvent)
	{
		m_pEvent->SetNotifyWindow(NULL, 0, 0);
		ref = m_pEvent->Release();
	}
	if(m_pControl)
		ref = m_pControl->Release();
	if(m_pBaseAudio)
		ref = m_pBaseAudio->Release();	
	if(m_pBaseVid)
		ref = m_pBaseVid->Release();
	if(m_pVidWin)
		ref = m_pVidWin->Release();

	// LAV Filters - app managed pointers.
	if(m_pLavAud)
	{
		m_pLavAud->SetSyncSource(NULL);
		m_pGraph->RemoveFilter(m_pLavAud);
		ref = m_pLavAud->Release();
	}
	if(m_pLavVid)
	{
		m_pLavVid->SetSyncSource(NULL);
		m_pGraph->RemoveFilter(m_pLavVid);
		ref = m_pLavVid->Release();
	}
	if(m_pLavSrc)
	{
		m_pLavSrc->SetSyncSource(NULL);
		m_pGraph->RemoveFilter(m_pLavSrc);
		ref = m_pLavSrc->Release();
	}

	// Release the graph.
	if(m_pGraph)
	{
		ref = m_pGraph->Release();
		if(ref != 0)
		{
			TRACE(_T("ERROR CPlayer CleanPlayback final ref count not zero.\n"));
		}
	}

	// Set ptrs etc.
	m_pGraph		 = NULL;
	m_pControl		 = NULL;
	m_pEvent		 = NULL;
	m_pVidWin		 = NULL;
	m_pBaseVid		 = NULL;
	m_pBaseAudio	 = NULL;
	m_pSeeking		 = NULL;
	m_redrawVid		 = FALSE;
	m_muted			 = FALSE;
	m_fsCurrent		 = State_Stopped;

	// LAV Filters.
	m_pLavSrc		= NULL;
	m_pLavVid		= NULL;
	m_pLavAud		= NULL;

	// Misc.
	m_rtTotalTime = 0;

	/*
	// Free LAV Filter dlls. Should not be neccessary.
	//
	// Per Process Explorer, 2124 k private bytes are held here
	// between playbacks. Khan.exe holds about 22440.
	//
	// NOTE: Will crash if this app's LAV DLL's have been registered
	// and DS Intelligent Connect has used them. For example, will 
	// occur if CDataView::OnTestDsPlayVideo() has been called.
	//
	HMODULE mod;
	BOOL free = FALSE;
	mod = GetModuleHandle(m_pathLavAud);
	if(mod)
	{
		TRACE(_T("Calling freeLibrary on LavAud\n"));
		free = AfxFreeLibrary(mod);
		if(!free)
		{
			TRACE(_T("Error FreeLibrary LavAud\n"));		
		}
	}
	mod = GetModuleHandle(m_pathLavVid);
	if(mod)
	{
		TRACE(_T("Calling freeLibrary on LavVid\n"));
		free = AfxFreeLibrary(mod);
		if(!free)
		{
			TRACE(_T("Error FreeLibrary LavVid\n"));		
		}
	}
	mod = GetModuleHandle(m_pathLavSrc);
	if(mod)
	{
		TRACE(_T("Calling freeLibrary on LavSrc\n"));
		free = AfxFreeLibrary(mod);
		if(!free)
		{
			TRACE(_T("Error FreeLibrary LavSrc\n"));		
		}
	}*/

	return ret;
}

///////////////////////////
// Control Video Playback.	

// Toggle pause / depause.
// Called by PausePlayback() and CDataView::ToggleVideoPlaybackPause().
BOOL CPlayer::TogglePlaybackPause()
{
    HRESULT hr = S_OK;

	// Not stopped.
    if(!m_pControl || m_fsCurrent == State_Stopped)
        return FALSE;

    // If running, pause
    if(m_fsCurrent == State_Running)
	{
		hr = m_pControl->Pause();
		if (FAILED(hr))
			return FALSE;
		m_fsCurrent = State_Paused;
	}
	// If paused, run.
    else if(m_fsCurrent == State_Paused)
	{
		// Resume playback
		hr = m_pControl->Run();
		if(FAILED(hr))
			return FALSE;
		m_fsCurrent = State_Running;
	}
	return TRUE;
}

// Pause the video. For deactivate.
// Called by...
void CPlayer::PausePlayback()
{
	// If not paused.
	if(m_fsCurrent == State_Running)
		TogglePlaybackPause();
}

// Set visability.
// Called from CDataView::OnDraw().
void CPlayer::PutVisible(BOOL hide)
{
	if(!m_pVidWin)
		return;

	// Hide
	if(hide)
		m_pVidWin->put_Visible(OAFALSE);
	// Show.
	else
		m_pVidWin->put_Visible(OATRUE);
}

// Mute demute volume.
// Called from CDataView::OnPlaybackMute().
void CPlayer::ToggleVolumeMute()
{
	if(m_pBaseAudio == NULL)
		return;

    HRESULT hr = S_OK;

	long lVolume;

    // Read current volume
    hr = m_pBaseAudio->get_Volume(&lVolume);
    if(hr == E_NOTIMPL)
        return;

    else if(FAILED(hr))
		return;

	// Demute.
	if(lVolume == DSBVOLUME_MIN) // -10000 muted
	{
		// Set to last volume
		LONG cur = DxVolFromPercent(m_curVolPercent);
		hr = m_pBaseAudio->put_Volume(cur);
		if(FAILED(hr))
			return;
		m_muted = FALSE;

	}
	// Mute
	else
	{
		// Set to mute volume
		hr = m_pBaseAudio->put_Volume(DSBVOLUME_MIN);
		if(FAILED(hr))
			return;
		m_muted = TRUE;
	}
}

// Step the volume up/down.
// Called from CDataView::OnMouseWheel().
BOOL CPlayer::StepVolume(BOOL stepUp, LONG & newVol)
{
    HRESULT hr = S_OK;

	if(m_muted || m_pBaseAudio == NULL)
		return FALSE;

	// Adjust percent - step by 4%.
	int percent = m_curVolPercent;
	percent += stepUp ? 4 : -4;
	if(percent > 100)
		percent = 100;
	if(percent < 0)
		percent = 0;

	// Get dx value.
	LONG vol = DxVolFromPercent(percent);

	// Set.
	hr = m_pBaseAudio->put_Volume(vol);
	if(FAILED(hr))
		return FALSE;

	m_curVolPercent = percent;
	newVol = percent;
	return TRUE;
}

// Stop playback.
void CPlayer::StopPlayback()
{
	CleanPlayback();
}

// Called from PlayVideoLav(), CDataView::OnSize().
BOOL CPlayer::SetVideoRect(CRect rect)
{
	TRACE(_T("CPlayer SetVideoRect setting rect.\n"));

	HRESULT hr;

	// Ensure a valid Video stream. Not used if only Audio.
	if(!m_pGraph		 ||
	   !m_pVidWin		 || 
	   !m_pBaseVid		 || 
	   rect.IsRectEmpty() )
	{
		TRACE(_T("CPlayer SetVideoRect no video stream.\n"));
		return FALSE;
	}

	double y = rect.Height();
	double x = rect.Width();

	// Get video size.
	LONG width, height;
	hr = m_pBaseVid->GetVideoSize(&width, &height);
	if(FAILED(hr))
		return CleanPlayback("ERROR - GetVideoSize failed.");

	// Get video size ratio.
	double ratio = (double)width / (double)height;

	// Maximize the video height.
	if( y * ratio <= x )
	{
		x -= y * ratio;
		x /= 2;

		// Set width.
		rect.DeflateRect((int)x, 0);
	}

	// Maximize the video width.
	else if( x / ratio <= y )
	{
		y -= x / ratio;
		y /= 2;

		// Set height.
		rect.DeflateRect(0, (int)y);
	}
	else
		return CleanPlayback("ERROR - SetVideoRect error.");

	// Set the video size.
	int horz = rect.Width();
	int vert = rect.Height();
	hr = m_pVidWin->SetWindowPosition(rect.left, rect.top, horz, vert);
	if(FAILED(hr))
		return CleanPlayback("ERROR SetVideoRect - SetWindowPosition failed.");

	// Ok
	return TRUE;
}

// Video playback is starting - prepare seeking.
void CPlayer::PrepareSeeking()
{
	if(m_fsCurrent != State_Running)
		return;

	// Set up playback time.
	m_rtTotalTime = 0;

	if(m_pSeeking) 
	{
		DWORD caps = AM_SEEKING_CanSeekAbsolute | 
					 AM_SEEKING_CanGetDuration; 
		if(m_pSeeking->CheckCapabilities(&caps) == S_OK)
		{
			// Get total vid time.
			m_pSeeking->GetDuration(&m_rtTotalTime);
		}
	}
}

// Get percent volume (0 to 100) from dx volume (0 to -10000).
int CPlayer::DxVolToPercent(LONG dxVol)
{
	// Calcs hacked From: 
	//	https://qt.gitorious.org/qt/qt/source/f2d89dee45e2f08e8e7613c18660c5c3e133c0ff:src/
	//	plugins/mediaservices/directshow/mediaplayer/directshowplayercontrol.cpp

	// Muted
	if(dxVol <= DSBVOLUME_MIN)	// -10000
		return 0;
	// Max
	if(dxVol >= DSBVOLUME_MAX)	// 0
		return 100;

	double d = 100.0 * pow( 10.0, (double)dxVol / 5000.0 );

	int percent = (int)d;
	if(percent > 100)
		percent = 100;
	if(percent < 0)
		percent = 0;

	return percent;
}

// Get dx volume (0 to -10000) from percent (0 to 100).
LONG CPlayer::DxVolFromPercent(int percent)
{
	if(percent <= 0)
		return DSBVOLUME_MIN; // -10000
	else if(percent >= 100)
		return DSBVOLUME_MAX; // 0

	double d1 = (double)percent / 100.0;
	double d2 = fabs( log10(d1) ) * -5000;

	// Verify
	LONG vol = (LONG)d2;
	if(vol < DSBVOLUME_MIN)
		return DSBVOLUME_MIN;
	if(vol > DSBVOLUME_MAX)
		return DSBVOLUME_MAX;

	return vol;
}

// Video Playback event callback.
// See PlayvideoLav().
//
// Called by the View when it is notified of a video 
// event via the WM_USER_VIDEOEVENT msg.
// See CDataView::OnVideoEvent().
//
// Init evCode with event code. View will manage event.
BOOL CPlayer::GetVideoEvent(LONG & evCode)
{
	if(!m_pEvent)
		return FALSE;

	LONG code, param1, param2;
	int nEvent = 0;

	// Check all events.
	while(m_pEvent && 
//		  ( SUCCEEDED(m_pEvent->GetEvent(&code, &param1, &param2, 0) ) ) )
		  ( SUCCEEDED(m_pEvent->GetEvent(&code, &param1, &param2, 100) ) ) ) // test timeout
	{
		// Clean event.
		m_pEvent->FreeEventParams(code, param1, param2);

		// dbg
		#ifdef _DEBUG
		// see: https://msdn.microsoft.com/en-us/library/windows/desktop/ff625885%28v=vs.85%29.aspx
		// Init param.
		evCode = code;
		TRACE(_T("VideoEvent Code = %d nEvent = %d\n"), code, nEvent);
		nEvent ++;
		switch(code)
		{
		case EC_CLOCK_CHANGED:			//	13, 0x0D
			TRACE(_T("VideoEvent Code = EC_CLOCK_CHANGED\n"));
			break;
		case EC_VMR_RENDERDEVICE_SET:	//	83, 0x53
			TRACE(_T("VideoEvent Code = EC_VMR_RENDERDEVICE_SET\n"));
			break;
		case EC_VIDEO_SIZE_CHANGED:		//	10, 0x0A
			TRACE(_T("VideoEvent Code = EC_VIDEO_SIZE_CHANGED\n"));
			break;
		case EC_PAUSED:					//	14, 0x0E
			TRACE(_T("VideoEvent Code = EC_PAUSED\n"));
			break;
		case EC_STREAM_ERROR_STOPPED:	//	6,	0x06
			TRACE(_T("VideoEvent Code = EC_STREAM_ERROR_STOPPED\n"));
			break;
		case EC_ERRORABORT:				//	3,	0x03
			TRACE(_T("VideoEvent Code = EC_ERRORABORT\n"));
			break;
		case EC_USERABORT:				//	2,	0x02
			TRACE(_T("VideoEvent Code = EC_USERABORT\n"));
			break;
		case EC_COMPLETE:				//  1,	0x01
			TRACE(_T("VideoEvent Code = EC_COMPLETE\n"));
			break;
		case EC_STARVATION:				// 23,	0x17
			TRACE(_T("VideoEvent Code = EC_STARVATION\n"));
			break;
		case EC_ACTIVATE:				// 19,	0x13
			TRACE(_T("VideoEvent Code = EC_ACTIVATE\n"));
			break;
		case EC_LOADSTATUS:				// 67,	0x43
			TRACE(_T("VideoEvent Code = EC_LOADSTATUS\n"));	// evcode.h
			break;
		case EC_OPENING_FILE:			// 10 ,	0x10
			TRACE(_T("VideoEvent Code = EC_OPENING_FILE\n"));
			break;
		default:
			break;
		}
		#endif // dbg

		// For now, only interested in playback completion.
		if(code == EC_COMPLETE || code == EC_USERABORT)
		{
			evCode = code;
			return TRUE;
		}
	} 
	return FALSE;
}

// ControlsBar Time slider moved.
// Moving param indicates user is dragging the time slider.
// NewPos param is the current Slider position.
// Seconds param is used by caller to update the slider.
BOOL CPlayer::PlaybackTimeSliderMoved(BOOL moving, const UINT & newPos, UINT & seconds)	
{
	// Ensure.
	if(!m_pControl || 
	   !m_pSeeking ||
	   m_fsCurrent == State_Stopped)
		return FALSE;

	HRESULT hr;

	// Calc the new ref time per newPos param.
	REFERENCE_TIME pos = (REFERENCE_TIME)newPos; // 0 - 1000
	REFERENCE_TIME rtNew = (m_rtTotalTime * pos) / 1000;

	// Calc the current seconds - will be used
	// by caller to set control.
	seconds = (long)(rtNew / 10000000);

	// If slider is being moved, pause the graph.
	if(moving)
	{
		// If running, pause
		if(m_fsCurrent == State_Running)
		{
			hr = m_pControl->Pause();
			if(FAILED(hr))
			{
				TRACE(_T("EROOR - CPlayer::OnPlaybackTime Pause failed\n"));
				return FALSE;
			}
			m_fsCurrent = State_Paused;
		}
		// Graph is paused - Update the position continuously.
		hr = m_pSeeking->SetPositions(&rtNew, 
									  AM_SEEKING_AbsolutePositioning,
									  NULL,
									  AM_SEEKING_NoPositioning);
		if(FAILED(hr))
		{
			TRACE(_T("EROOR - CPlayer::OnPlaybackTime SetPosition 1 failed\n"));
			return FALSE;
		}
	}

	// Else slider has been released - de-pause if paused.
	else if(m_fsCurrent == State_Paused)
	{
		// Set the final position.
		hr = m_pSeeking->SetPositions(&rtNew, 
									  AM_SEEKING_AbsolutePositioning,
									  NULL,
									  AM_SEEKING_NoPositioning);
		if(FAILED(hr))
		{
			TRACE(_T("EROOR - CPlayer::OnPlaybackTime SetPosition 2 failed\n"));
			return FALSE;
		}

		// De-pause playback.
		hr = m_pControl->Run();
		if(FAILED(hr))
		{
			TRACE(_T("EROOR - CPlayer::OnPlaybackTime Run failed\n"));
			return FALSE;
		}
		m_fsCurrent = State_Running;
	}
	return TRUE;
}

// ControlsBar vol slider moved.
// NewPos param is percent, 0 - 100.
BOOL CPlayer::PlaybackVolSliderMoved(int newPos)
{
    HRESULT hr = S_OK;

	if(m_muted || m_pBaseAudio == NULL)
		return FALSE;

	// Adjust percent.
	int percent = (int)newPos;
	if(percent > 100)
		percent = 100;
	if(percent < 0)
		percent = 0;

	// Get dx value.
	LONG vol = DxVolFromPercent(percent);

	// Set.
	hr = m_pBaseAudio->put_Volume(vol);
	if(FAILED(hr))
		return FALSE;
	m_curVolPercent = percent;

	return TRUE;
}


// Get the current Video Time:
//	position (0 - 10000)
//	seconds  (0 - 1190 hrs).
BOOL CPlayer::PlaybackGetCurTime(UINT & position, UINT & seconds) 
{
	// If playback not stopped / paused.
	if(m_pSeeking && m_fsCurrent == State_Running)
	{
		REFERENCE_TIME pos, sec, rtNow = 0;
		HRESULT hr = m_pSeeking->GetCurrentPosition(&rtNow);
		if(SUCCEEDED(hr))
		{
			pos = rtNow * 1000 / m_rtTotalTime;	// 1000 divs
			sec = rtNow / 10000000;				// cur play seconds
			position = (UINT) pos;
			seconds  = (UINT) sec;
			return TRUE;			
		}
	}
	return FALSE;
}
/////////////////////////////////////////////////////////
// Helpers

// Called from CPlayer::CleanPlayback() in debug only.
// Show filters used in the graph.
// From msdn: Enumerating Filters
HRESULT CPlayer::EnumFilters()
{
	TRACE(_T("\n"));
	TRACE(_T("Entering CPlayer EnumFilters\n"));

	IEnumFilters *pEnum = NULL;
	IBaseFilter *pFilter;
	ULONG cFetched;

	HRESULT hr = m_pGraph->EnumFilters(&pEnum);
	if (FAILED(hr)) 
	return hr;

	while(pEnum->Next(1, &pFilter, &cFetched) == S_OK)
	{
		FILTER_INFO FilterInfo;
		hr = pFilter->QueryFilterInfo(&FilterInfo);
		if (FAILED(hr))
		{
			TRACE(_T("Could not get the filter info\n"));
			continue;  // Maybe the next one will work.
		}
		TRACE(_T("Using filter: %s\n"), FilterInfo.achName);

		// The FILTER_INFO structure holds a pointer to the Filter Graph
		// Manager, with a reference count that must be released.
		if (FilterInfo.pGraph != NULL)
		{
			FilterInfo.pGraph->Release();
		}
		pFilter->Release();
	}
	TRACE(_T("\n"));

	pEnum->Release();
	return S_OK;
}








